local MAJOR = "StarLibWidgetText-1.0" 
local MINOR = 1

assert(LibStub, MAJOR.." requires LibStub") 
local WidgetText = LibStub:NewLibrary(MAJOR, MINOR)
assert(WidgetText, "LibStub.NewLibrary failed")
local LibProperty = LibStub("StarLibProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibWidget = LibStub("StarLibWidget-1.0", true)
assert(LibWidget, MAJOR .. " requires StarLibWidget-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")
local LibRGBA = LibStub("StarLibRGBA-1.0", true)
assert(LibRGBA, MAJOR .. " requires StarLibRGBA-1.0")


local PINGPONGWAIT = 2
local ALIGN_LEFT, ALIGN_CENTER, ALIGN_RIGHT, ALIGN_MARQUEE, ALIGN_AUTOMATIC, ALIGN_PINGPONG = 1, 2, 3, 4, 5, 6
local aligns = {L = 1, C = 2, R = 3, M = 4, A = 5, P = 6}
local alignments = {"Left", "Center", "Right", "Marquee", "Automatic", "Pingpong"}
local alignmentsDict = {Left = 1, Center = 2, Right = 3, Marquee = 4, Automatic = 5, Pingpong = 6}

local SCROLL_RIGHT, SCROLL_LEFT = 1, 2
local scrolls = {"Right", "Left"}
local scrollsDict = {Right = 1, Left = 2}

local map = {value = "Value", prefix = "Prefix", postfix = "Postfix", precision = "Precision", align = "Alignment", update = "Update", speed = "Scroll Speed", direction ="Direction", cols = "Columns"}

local textUpdate, scrollUpdate
local pool = setmetatable({}, {__mode = "k"})

if not WidgetText.__index then
	WidgetText.__index = WidgetText
end

--[[
local function resizeText(str, cols)
	if strlen(str) < cols then
		for i = strlen(str) + 1, cols do
			str = str .. " "
		end
	end
	return str:sub(1, cols)
end

local function insertChar(text1, pos1, text2, pos2)
	local char = text2:sub(pos2, pos2)
	
	local left = text1:sub(1, pos1 - 1)
	local right = text1:sub(pos1 + 1)
	
	return left .. char .. right
end
]]

WidgetText.defaults = {
	precision = 0xbabe,
	align = ALIGN_LEFT,
	update = 0,
	speed = 500,
	direction = SCROLL_RIGHT,
	cols = 40
}

-- Create a new StarLibWidgetText object
-- @name WidgetText.New
-- @usage WidgetText:New(visitor, name, config, row, col, layer, fontString, env, errorLevel)
-- @param visitor An StarLibCore-1.0 object
-- @param config This widget's settings
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param env The environment for the script
-- @param config The settings. You can provide these fields: string, prefix, postfix, precision, align, update, scroll, speed, direction, cols, offset, str, update, speed. Check the source for further details. Some fields have defaults.
-- @param self.errorLevel The self.errorLevel for this object
-- @return A new StarLibWidgetText object

function WidgetText:New(visitor, name, config, row, col, layer, environment, errorLevel, callback, data) 

	assert(environment, "Please provide the marquee with an environment")
	assert(config, "Please provide the marquee with a config")
	assert(config.name, "Please provide the marquee with a name")
	assert(config.string, config.name .. ": Please provide the marquee with a string")
	
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end
	
	obj.error = LibError:New(MAJOR, errorLevel)
	
	setmetatable(obj, self)

	obj.visitor = visitor
	
	obj.errorLevel = errorLevel or 3

	obj.value = LibProperty:New(visitor, config.name .. " string", config.string, "", errorLevel) -- text of marquee
	obj.prefix = LibProperty:New(visitor, config.name .. " prefix", config.prefix, "", errorLevel) -- label on the left side
	obj.postfix = LibProperty:New(visitor, config.name .. " postfix", config.postfix, "", errorLevel) -- label on right side
	obj.precision = config.precision or self.defaults.precision -- number of digits after the decimal point
	obj.align = aligns[config.align] or self.defaults.align -- alignment: left, center, right, marquee, automatic
	obj.update = config.update or self.defaults.update -- update interval
	obj.scroll = 0 -- marquee starting point
	obj.speed = config.speed or self.defaults.speed -- marquee scrolling speed
	obj.direction = config.direction or self.defaults.direction -- pingpong direction, 0 = right, 1 = left
	obj.cols = config.length or self.defaults.cols -- number of colums in marquee
	obj.offset = 0 -- increment by pixel
	obj.string = "" -- formatted value
	obj.timer = LibTimer:New("WidgetText.timer", obj.update, true, textUpdate, obj)
	obj.textTimer = LibTimer:New("WidgetText.textTimer", obj.speed, true, textScroll, obj)
	obj.callback = callback
	obj.data = data
		
	obj.options = {}

	obj.widget = LibWidget:New(obj, visitor, name, config, row, col, layer, {text=true, rc=true}, env, errorLevel)
	
	obj.color = LibRGBA:New()
	obj.color_valid = obj.widget:WidgetColor(config, "color", obj.color);	
		
	return obj	
end

-- Delete a StarLibWidgetText object
-- @name StarLibWidgetText.Del
-- @usage StarLibWidgetText:Del([marq])
-- @param marq A StarLibWidgetText object
function WidgetText:Del(marq)
	if not marq then
		marq = self
	end
	marq:Stop()
	pool[marq] = true
	marq.widget:Del()
end

-- Start a StarLibWidgetText
-- @name StarLibWidgetText.Start
-- @usage object:Start()
-- @return Nothing
function WidgetText:Start()
	self.error:Print("WidgetText:Start", 3)
	textUpdate(self)	
	textScroll(self)
	if self.update ~= 0 then
		self.timer:Start()
		self.textTimer:Start()
		StarTip:Print("start timers")
	end
end

-- Stop a StarLibWidgetText
-- @name StarLibWidgetText.Stop
-- @usage object:Stop()
-- @return Nothing
function WidgetText:Stop()
	self.error:Print("WidgetText:Stop", 3)
	self.timer:Stop()
	self.textTimer:Stop()
end

-- Print text on the font string
-- @name StarLibWidgetText.Draw
-- @param text The text to print. Feel free to hook this.
-- @return Nothing
function WidgetText:Draw()
	if type(self.callback) == "function" then
		self.callback(self, self.data)
	end
end

function textScroll(self)
	if not self.prefix or not self.postfix then return end
	
	self.count = (self.count or 0) + 1
	local pre = self.prefix:P2S()
	local post = self.postfix:P2S()

	local str = self.string
	
	local num, len, width, pad = 0, strlen(str), self.cols - strlen(pre) - strlen(post), 0
	local srcPtr, dstPtr = 0, 0
	local src = ""
	local dst = ""
	
    if width < 0 then
        width = 0
	end
	
	if self.align == ALIGN_LEFT then
		pad = 0
	elseif self.align == ALIGN_CENTER then
		pad = (width - len) / 2
		if pad < 0 then
			pad = 0
		end
	elseif self.align == ALIGN_RIGHT then
		pad = width - len
		if pad < 0 then
			pad = 0
		end
	elseif self.align == ALIGN_AUTOMATIC then
		if len <= width then
			pad = 0
		end
	elseif self.align == ALIGN_MARQUEE then
		pad = width - self.scroll
		self.scroll = self.scroll + 1
		if self.scroll >= width + len then
			self.scroll = self.cols
		end
	elseif self.align == ALIGN_PINGPONG then
		if len <= width then
			pad = (width - len) / 2
		else
			if self.direction == SCROLL_RIGHT then
				self.scroll = self.scroll + 1
			else
				self.scroll = self.scroll - 1
			end
			
			pad = 0 - scroll_;
			
			if pad < 0 - (len - width) then
				if self.delay < 1 then
					self.direction = SCROLL_LEFT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll - PINGPONGWAIT
				end
				pad = 0 - (len - width)
			elseif pad > 0 then
				if self.delay < 1 then
					self.direction = SCROLL_RIGHT
					self.delay = PINGPONGWAIT
					self.scroll = self.scroll + PINGPOINGWAIT
				end
				pad = 0
			else
				pad = 0
			end
		end
	else
		pad = 0
		self.error:Print("No alignment specified")
	end

	dst = LibUtils:resizeText(dst, self.cols)
	
    dstPtr = 0;
    num = 0;

    -- /* process prefix */
    src = pre;
    while (num < self.cols) do
        if (srcPtr == strlen(src)) then
            break
		end
		dst = LibUtils:replaceText(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end

    src = str;
    srcPtr = 0;
	
	local offset = pad
	
    if(offset < 0) then
        offset = 0;
	end
	
    -- /* wrap around on the beginning */
    while (pad > 0 and num < self.cols) do
        if(self.align == ALIGN_MARQUEE) then
			dst = LibUtils:replaceText(dst, dstPtr, src, (strlen(src) - offset) + srcPtr)
            --dst[dstPtr] = src[(strlen(src) - offset) + srcPtr];
			dstPtr = dstPtr + 1
			srcPtr = srcPtr + 1
        else
            dst[dstPtr] = ' ';
		end
		dstPtr = dstPtr + 1
        num = num + 1
        pad = pad - 1
    end

    --/* skip src chars (marquee) */
	local tmp = src
    while (pad < 0 and tmp ~= "") do
        src = tmp:sub(1); 
        tmp = src;
        pad = pad + 1
    end


    --/* copy content */
    while (num < self.cols) do
        if (srcPtr >= strlen(src)) then
            break;
		end
		dst = LibUtils:replaceText(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end
	
    -- wrap around on end 
    src = post;
    len = strlen(src)
    srcPtr = 0;
    if(num < self.cols - len and self.align == ALIGN_MARQUEE) then
		dst = LibUtils:replaceText(dst, dstPtr, "*", 1)
        --dst[dstPtr] = '*';
		dstPtr = dstPtr + 1
        num = num + 1
    end
    while (num < self.cols - len) do
        if(self.align == ALIGN_MARQUEE) then
			dst = LibUtils:replaceText(dst, dstPtr, src, srcPtr)
            --dst[dstPtr] = str[srcPtr];
			dstPtr = dstPtr + 1
			srcPtr = srcPtr + 1
        else
			dst = LibUtils:replaceText(dst, dstPtr, ' ', 1)
            --dst[dstPtr] = ' ';
			dstPtr = dstPtr + 1
		end
        num = num + 1;
    end

    srcPtr = 0;

    --/* process postfix */
    while (num < self.cols) do
        if (srcPtr >= strlen(src)) then
            break;
		end
		dst = LibUtils:replaceText(dst, dstPtr, src, srcPtr)
        --dst[dstPtr] = src[srcPtr];
		dstPtr = dstPtr + 1
		srcPtr = srcPtr + 1
        num = num + 1
    end
	
	self.buffer = dst
	
	self:Draw(self.buffer)
end


function textUpdate(self)
	local str
	local update = 0
	
	assert(self.value, "WidgetText has no value")
	if not self.prefix then return self.error:Print("WidgetText needs a prefix")end
	if not self.postfix then return self.error:Print("WidgetText needs a postfix") end
	
	update = update + self.prefix:Eval()
	update = update + self.postfix:Eval()
	
	self.value:Eval()
	
    -- /* str or number? */
    if (self.precision == 0xBABE) then
        str = self.value:P2S();
    else
        local number = self.value:P2N();
        local width = self.cols - strlen(self.prefix:P2S()) - strlen(self.postfix:P2S());
        local precision = self.precision;
        --[[/* print zero bytes so we can specify NULL as target  */
        /* and get the length of the resulting str */]]
		local text = ("%.*f"):format(precision, number)
		local size = strlen(text)
        --/* number does not fit into field width: try to reduce precision */
        if (width < 0) then
            width = 0;
		end
        if (size > width and precision > 0) then
            local delta = size - width;
            if (delta > precision) then
                delta = precision;
			end
            precision = precision - delta;
            size = size - delta;
            --/* zero precision: omit decimal point, too */
            if (precision == 0) then
                size = size - 1
			end
        end
        ---/* number still doesn't fit: display '*****'  */
        if (size > width) then
            str.resize(width);
            for i = 0, width do
                str[i] = '*';
			end
        else
            str = text
        end
    end

    if str == "" or str ~= self.string then 
        update = update + 1;
        self.string = str;
    end

    --/* something has changed and should be updated */
    if (update > 0) then

		
        -- /* Init pingpong scroller. start scrolling left (wrong way) to get a delay */
        if (self.align == ALIGN_PINGPONG) then
            self.direction = 0;
            self.delay = PINGPONGWAIT;
        end
		--[[
        /* if there's a marquee scroller active, it has its own */
        /* update callback timer, so we do nothing here; otherwise */
        /* we simply call this scroll callback directly */
		]]
        if (self.align ~= ALIGN_MARQUEE or self.align ~= ALIGN_AUTOMATIC or self.align ~= ALIGN_PINGPONG) then
            textScroll(self)
			return false
        end

    end

	return true
end

function WidgetText:BuildOptions(config, db)
	self.options = {[self.name] = {
		name = config.name,
		type = "group",
		args = { }
	}}
	for k, v in pairs(config) do
		t[self.name].args.widget = {
			value = {
				name = "Value",
				type = "input",
				multiline = true,
				get = function()
					return db.value or self.value
				end,
				set = function(info, v)
					db.value = v
					if self.value then
						self.value:Del()
						self.value = nil
					end
					self.value = LibProperty:New(self.visitor, self.name, v, "", self.errorLevel) -- text of marquee
					
					self.value = v
				end
			},
			prefix = {
				name = "Prefix",
				type = "input",
				multiline = true,
				get = function()
					return db.prefix or self.prefix
				end,
				set = function(info, v)
					db.prefix = v
					if self.prefix then
						self.prefix:Del()
						self.prefix = nil
					end
					self.prefix = LibProperty:New(self.visitor, self.name, v, "", self.errorLevel) -- text of marquee
				end
			},
			postfix = {
				name = "Postfix",
				type = "input",
				multiline = true,
				get = function()
					return db.postfix or self.postfix
				end,
				set = function(info, v)
					db.postfix = v
					if self.postfix then
						self.postfix:Del()
						self.postfix = nil
					end
					self.postfix = LibProperty:New(self.visitor, self.name, v, "", self.errorLevel) -- text of marquee
				end
			},
			precision = {
				name = "Precision",
				type = "input",
				pattern = "%d",
				get = function()
					return db.precision or self.precision
				end,
				set = function(info, v)
					db.precision = v
					self.precision = v
				end
			},
			align = {
				name = "Alignment",
				type = "select",
				values = alignments,
				get = function()
					return alignmentDict[db.align] or self.align
				end,
				set = function(info, v)
					db.align = alignments[v]
					self.align = v
				end,
				order = 5
			},
			update = {
				name = "Update",
				type = "input",
				pattern = "%d",
				get = function() 
					return db.update or self.update
				end,
				set = function(info, v)
					db.update = v
					self.update = v
					if self.timer then
						self.timer:Del()
						self.timer = nil
					end
					self.timer = LibTimer:New("WidgetText.timer", self.update, true, textUpdate, self)
				end
			},
			speed = {
				name = "Scroll Speed",
				type = "input",
				pattern = "%d",
				get = function()
					return db.speed or self.speed
				end,
				set = function(info, v)
					db.speed = v
					self.speed = v
					if self.textTimer then
						self.textTimer:Stop()
						self.textTimer:Del()
						self.textTimer = nil
					end
					self.textTimer = LibTimer:New("WidgetText.textTimer", self.speed, true, textScroll, self)
				end
			},
			direction = {
				name = "Direction",
				type = "select",
				values = directions,
				get = function()
					return scrollsDict[db.direction] or self.direction
				end,
				set = function(info, v)
					db.direction = scrolls[v]
					self.direction = v
				end
			},
			cols = {
				name = "Columns",
				type = "input",
				pattern = "%d",
				get = function()
					return db.cols or self.cols
				end,
				set = function(info, v)
					db.cols = v
					self.cols = v
				end
			},
			color = {
				name = "Color",
				type = "input",
				pattern = "%d",
				get = function()
					return db.color.r or self.color.R, db.color.g or self.color.G, db.color.b or self.color.B, db.color.a or self.color.A
				end,
				set = function(info, r, g, b, a)
					db.color.r = r
					self.color.R = r
					db.color.g = g
					self.color.G = g
					db.color.b = b
					self.color.B = b
					db.color.a = a
					self.color.A = a
				end,
			}
		}
	end
end
